#!/usr/bin/env python

import sys
from logging import Logger as PythonLogger

#  --- Copied from logging module ---
if hasattr(sys, '_getframe'):
    def currentframe():
        return sys._getframe(3)
else:  # pragma: no cover
    def currentframe():
        """Return the frame object for the caller's stack frame."""
        try:
            raise Exception
        except Exception:
            return sys.exc_info()[2].tb_frame.f_back


#  --- Copied from logging module ---


class KomoLogger(PythonLogger):
    def __init__(self, name: str):
        super().__init__(name)



__version__ = '0.1.0'

import logging

from komolibs.logger.struct_logger import StructLogRecord, StructLogger

STRUCT_LOGGER_SET = False
DEV_STRATEGY_PREFIX = "dev"
_prefix_path = None

# Do not raise exceptions during log handling
logging.setLogRecordFactory(StructLogRecord)
logging.setLoggerClass(StructLogger)


from typing import Any

from influxdb_client import InfluxDBClient, Point
from influxdb_client.client.write_api import SYNCHRONOUS


class InfluxData:
    def __init__(self, url: str, org: str, token: str, bucket: str):
        self._client = InfluxDBClient(url=url, token=token)
        self._write_api = self._client.write_api(write_options=SYNCHRONOUS)
        self._org = org
        self._bucket = bucket

    @property
    def bucket(self) -> str:
        return self._bucket

    def write(self, point: Point):
        self._write_api.write(self._bucket, self._org, point)

    def batch(self, sequence: Any):
        self._write_api.write(self._bucket, self._org, sequence)

    def query(self, query: str):
        return self._client.query_api().query(query, org=self._org)

    def is_node_healthy(self):
        status = self._client.health().status
        print(f"HEALTH {self._client.ready()}")
        return True  # TODO



import codecs
import csv
import io
import json
from typing import Any

from komolibs.aws.s3.s3_base import S3Base


class Download(S3Base):
    def __init__(self):
        super().__init__()

    def download(self, bucket: str, file_name: str) -> Any:
        try:
            s3_object = self.resource.Object(bucket, file_name)
            data = io.BytesIO()
            s3_object.download_fileobj(data)

            # object is now a bytes string, Converting it to a dict:
            return json.loads(data.getvalue().decode("utf-8"))

        except Exception as e:
            raise e

    def download_csv(self, bucket: str, file_name: str) -> csv.DictReader:
        s3_object = self.client.get_object(Bucket=bucket, Key=file_name)

        return csv.DictReader(codecs.getreader("utf-8")(s3_object["Body"]))



from aioredis import create_redis


async def check_redis_connection(url: str,
                                 port: int,
                                 password: str,
                                 db: int = 0,
                                 ssl: bool = True):
    val_in: str = 'value'
    val_out: any = None
    try:
        # Redis client bound to single connection (no auto reconnection).
        redis = await create_redis((url, port), db=db, password=password, ssl=ssl)

        await redis.set('test-connection-key', val_in)
        val_out = await redis.get('test-connection-key')

        # gracefully closing underlying connection
        redis.close()
        await redis.wait_closed()

    except Exception as e:
        pass

    if val_out.decode("utf-8") == val_in:
        return True
    else:
        raise Exception("Redis connection test failed. Check redis server status. ")



import json

from komolibs.aws.s3.s3_base import S3Base, S3Result


class Upload(S3Base):
    def __init__(self):
        super().__init__()

    def upload(self, data: dict, file_name: str, bucket: str) -> S3Result:
        try:
            s3_object = self.resource.Object(bucket, file_name)
            text = json.dumps(data)
            result = s3_object.put(Body=text)
            response = result.get('ResponseMetadata')
            if response.get('HTTPStatusCode') in [200, 201]:
                self.logger().info('File Uploaded Successfully. ')
                return S3Result.SUCCESS
            else:
                self.logger().error(f'File Not Uploaded. {response.get("HTTPStatusCode")}')
                return S3Result.FAILURE
        except Exception as e:
            raise e


import asyncio
import logging
from asyncio import Task
from typing import Optional

from aioredis import Redis, create_redis

from komolibs.core.utils.async_utils import safe_ensure_future
from komolibs.logger import KomoLogger


class Publisher:
    publisher_logger: Optional[KomoLogger] = None
    _shared_instance: "Publisher" = None

    @classmethod
    def get_instance(cls,
                     url: str,
                     port: int,
                     password: str,
                     db: int = 0,
                     ssl: bool = True,
                     channel: Optional[str] = "*") -> "Publisher":
        if cls._shared_instance is None:
            cls._shared_instance = Publisher(url=url, port=port, password=password, db=db, ssl=ssl, channel=channel)
        return cls._shared_instance

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls.publisher_logger is None:
            cls.publisher_logger = logging.getLogger(__name__)
        return cls.publisher_logger

    def __init__(self,
                 url: str,
                 port: int,
                 password: str,
                 db: int = 0,
                 ssl: bool = True,
                 channel: Optional[str] = "*"):
        self._url: str = url
        self._port = port
        self._password = password
        self._db = db
        self._ssl =ssl
        self._channel: Optional[str] = channel
        self._redis: Optional[Redis] = None
        self._ready: bool = False

        self._publish_message_loop_task: Optional[Task] = None
        self._publish_stream: asyncio.Queue = asyncio.Queue()

    @property
    def ready(self):
        return self._ready

    @property
    def publish_stream(self) -> asyncio.Queue:
        return self._publish_stream

    async def publish(self, message: dict):
        self.publish_stream.put_nowait(message)

    async def start(self):
        self._redis = await create_redis((self._url, self._port), db=self._db, password=self._password, ssl=self._ssl)
        # await self._redis.set("MYKEY", "HELLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLOOOOOOOOOOOOOOOOOOOOOOO")
        # key = await self._redis.get("MYKEY")
        # print(key)
        self._publish_message_loop_task = safe_ensure_future(self.publish_message_loop())
        self._ready = True

    async def stop(self):
        self._redis.close()
        await self._redis.wait_closed()
        if self._publish_message_loop_task is not None:
            self._publish_message_loop_task.cancel()
            self._publish_message_loop_task = None

    async def publish_message_loop(self):
        message = None
        try:
            while True:
                try:
                    message = await self._publish_stream.get()
                    assert type(message) is dict, "Only dictionary type is processed by this publisher."
                    assert self._channel is not None, f"Please supply valid publication channel. {self._channel} is not valid. "
                    await self._redis.publish_json(self._channel, message)
                    await asyncio.sleep(0.1)
                except AssertionError as e:
                    self.logger().error(f"{e}. {type(message)}")
        except asyncio.CancelledError:
            raise
        except Exception:
            raise
        finally:
            await self.stop()



import asyncio
import logging
import time
import inspect


async def safe_wrapper(c):
    try:
        return await c
    except asyncio.CancelledError:
        raise
    except Exception as e:
        logging.getLogger(__name__).error(f"Unhandled error in background task: {str(e)}", exc_info=True)


def safe_ensure_future(coro, *args, **kwargs):
    return asyncio.ensure_future(safe_wrapper(coro), *args, **kwargs)


async def safe_gather(*args, **kwargs):
    try:
        return await asyncio.gather(*args, **kwargs)
    except Exception as e:
        logging.getLogger(__name__).debug(f"Unhandled error in background task: {str(e)}", exc_info=True)
        raise


async def wait_til(condition_func, timeout=10):
    start_time = time.perf_counter()
    while True:
        if condition_func():
            return
        elif time.perf_counter() - start_time > timeout:
            raise Exception(f"{inspect.getsource(condition_func).strip()} condition is never met. Time out reached.")
        else:
            await asyncio.sleep(0.1)


async def run_command(*args):
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE)
    stdout, stderr = await process.communicate()
    return stdout.decode().strip()


import logging
from enum import Enum
from typing import Optional

import boto3

from komolibs.logger import KomoLogger


class S3Result(Enum):
    SUCCESS = 0
    FAILURE = 1


class S3Base:
    s3_base_logger: Optional[KomoLogger] = None

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls.s3_base_logger is None:
            cls.s3_base_logger = logging.getLogger(__name__)
        return cls.s3_base_logger

    def __init__(self):
        self._resource = boto3.Session()

    @property
    def resource(self):
        return self._resource.resource('s3')

    @property
    def client(self):
        return boto3.client("s3")



import asyncio
import logging
from asyncio import Task
from typing import Optional

from aioredis import Redis, create_redis

from komolibs.core.utils.async_utils import safe_ensure_future
from komolibs.logger import KomoLogger


class CacheWriter:
    cache_writer_logger: Optional[KomoLogger] = None
    _shared_instance: "CacheWriter" = None

    @classmethod
    def get_instance(cls, url: str, record_key: str) -> "CacheWriter":
        if cls._shared_instance is None:
            cls._shared_instance = CacheWriter(url=url, record_key=record_key)
        return cls._shared_instance

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls.cache_writer_logger is None:
            cls.cache_writer_logger = logging.getLogger(__name__)
        return cls.cache_writer_logger

    def __init__(self, url: str, record_key: str):
        self._url: str = url
        self._redis: Optional[Redis] = None
        self._record_key = record_key
        self._ready: bool = False

        self._cache_record_task: Optional[Task] = None
        self._cache_queue: asyncio.Queue = asyncio.Queue()

    @property
    def ready(self):
        return self._ready

    @property
    def cache_queue(self) -> asyncio.Queue:
        return self._cache_queue

    async def start(self):
        self._redis = await create_redis(str(self._url))
        self._cache_record_task = safe_ensure_future(self._cache_record_loop())
        self._ready = True

    async def write(self, message: str):
        self._cache_queue.put_nowait(message)

    async def stop(self):
        if self._cache_record_task is not None:
            self._cache_record_task.cancel()
            self._cache_record_task = None
        await asyncio.sleep(1)
        self.logger().info(f"Stopping cache writer redis connections.")
        self._redis.close()
        await self._redis.wait_closed()
        self.logger().info(f"Finished stopping cache writer redis connection. ")

    async def _cache_record_loop(self):
        self.logger().info(f"Starting cache record loop.")
        while True:
            try:
                record = await self._cache_queue.get()
                await self._redis.set(self._record_key, record)
            except asyncio.CancelledError:
                self.logger().error("Got CancelledError. ")
                break
            except Exception as e:
                raise e



import asyncio
import logging
from asyncio import Task
from typing import Optional, Any

import aioredis
from aioredis import create_redis

from komolibs.core.utils.async_utils import safe_ensure_future
from komolibs.logger import KomoLogger

STOPWORD = "STOP"


class Subscriber:
    subscriber_logger: Optional[KomoLogger] = None
    _shared_instance: "Subscriber" = None

    @classmethod
    def get_instance(cls,
                     url: str,
                     port: int,
                     password: str,
                     db: int = 0,
                     ssl: bool = True,
                     channel: Optional[str] = "*",
                     output: asyncio.Queue = asyncio.Queue()) -> "Subscriber":
        if cls._shared_instance is None:
            cls._shared_instance = Subscriber(url=url, port=port, password=password, db=db, ssl=ssl, channel=channel, output=output)
        return cls._shared_instance

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls.subscriber_logger is None:
            cls.subscriber_logger = logging.getLogger(__name__)
        return cls.subscriber_logger

    def __init__(self,
                 url: str,
                 port: int,
                 password: str,
                 db: int = 0,
                 ssl: bool = True,
                 channel: Optional[str] = "*",
                 output: asyncio.Queue = asyncio.Queue()):
        self._url: str = url
        self._port = port
        self._password = password
        self._db = db
        self._ssl = ssl
        self._channel: Optional[str] = channel
        self._redis: Optional[Any] = None
        self._pubsub: Optional[Any] = None
        self._ready: bool = False

        self._subscribe_task: Optional[Task] = None
        self._message_stream: asyncio.Queue = output

    @property
    def ready(self):
        return self._ready

    async def start(self):
        try:
            self._redis = await create_redis((self._url, self._port), db=self._db, password=self._password, ssl=self._ssl)
            res = await self._redis.subscribe(self._channel)
            self._pubsub = res[0]

            self._subscribe_task = safe_ensure_future(self.redis_subscriber_loop())
            self._ready = True
        except Exception as e:
            self._ready = False
            raise e

    async def stop(self):
        self._redis.close()
        await self._redis.wait_closed()
        if self._subscribe_task is not None:
            self._subscribe_task.cancel()
            self._subscribe_task = None

    @property
    def message_stream(self) -> asyncio.Queue:
        return self._message_stream

    async def redis_subscriber_loop(self):
        while True:
            try:
                while await self._pubsub.wait_message():
                    msg = await self._pubsub.get_json()
                    self.message_stream.put_nowait(msg)
            except asyncio.TimeoutError:
                raise
            except Exception:
                raise
            finally:
                await self.stop()



import logging
from decimal import Decimal
from enum import Enum
from logging import (
    DEBUG,
    INFO,
    WARNING,
    ERROR,
    CRITICAL
)

from .logger import KomoLogger


def log_encoder(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, Enum):
        return str(obj)
    raise TypeError("Object of type '%s' is not JSON serializable" % type(obj).__name__)


__all__ = [
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "KomoLogger",
    "log_encoder"
]
logging.setLoggerClass(KomoLogger)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['komolibs',
 'komolibs.aws',
 'komolibs.aws.s3',
 'komolibs.caching',
 'komolibs.core',
 'komolibs.core.utils',
 'komolibs.db',
 'komolibs.db.influx',
 'komolibs.logger',
 'komolibs.messaging',
 'komolibs.pubsub']

package_data = \
{'': ['*']}

install_requires = \
['aioredis>=1.3.1,<2.0.0',
 'boto3>=1.18.27,<2.0.0',
 'influxdb-client>=1.19.0,<2.0.0',
 'pendulum>=2.1.2,<3.0.0']

setup_kwargs = {
    'name': 'komolibs',
    'version': '0.1.410',
    'description': '',
    'long_description': None,
    'author': 'Makhosonke Morafo',
    'author_email': 'makhosonke@komokun.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import Optional

from aioredis import Redis, create_redis

from komolibs.logger import KomoLogger


class CacheReader:
    cache_reader_logger: Optional[KomoLogger] = None
    _shared_instance: "CacheReader" = None

    @classmethod
    def get_instance(cls, url: str, record_key: str) -> "CacheReader":
        if cls._shared_instance is None:
            cls._shared_instance = CacheReader(url=url, record_key=record_key)
        return cls._shared_instance

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls.cache_reader_logger is None:
            cls.cache_reader_logger = logging.getLogger(__name__)
        return cls.cache_reader_logger

    def __init__(self, url: str, record_key: str):
        self._url: str = url
        self._redis: Optional[Redis] = None
        self._record_key = record_key
        self._ready: bool = False

    @property
    def ready(self):
        return self._ready

    async def start(self):
        self._redis = await create_redis(str(self._url))
        self._ready = True

    async def stop(self):
        self.logger().info(f"Stopping cache reader redis connections.")
        self._redis.close()
        await self._redis.wait_closed()
        self.logger().info(f"Finished stopping cache reader redis connection. ")

    async def read(self):
        try:
            return await self._redis.get(self._record_key)
        except Exception as e:
            raise e

